# frozen_string_literal: trues

require 'timeout'

module Hermes
  # HousingIntention model
  class HousingApplication
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::Status
    include Mongoid::Attributes::Dynamic
    
    include Hermes::Concerns::SopraXmlSender
    field :applicated_at, type: DateTime
    field :effort_rate, type: Float
    field :epci, type: Integer
    field :is_active, type: Boolean, default: true
    field :max_date, type: DateTime
    field :visit_conducted_date, type: DateTime
    field :date_event, type: Date
    field :last_status_update_date, type: Date
    field :lessor_event_informations, type: String
    field :sopra_code, type: String
    field :als_id, type: String
    field :rank, type: Integer
    
    field :profile_score, type: Integer
    field :eligible_position, type: Integer
    field :applicated_position, type: Integer
    field :effort_rate, type: Float

    field :cotations, type: Hash
    field :business_cotation, type: Float

    field :requester_firstname, type: String
    field :requester_lastname, type: String
    field :requester_email, type: String
    field :requester_birth_date, type: Date
    field :applicant_used_for_siret_firstname, type: String
    field :applicant_used_for_siret_lastname, type: String
    field :applicant_used_for_siret_email, type: String
    field :applicant_used_for_siret_birth_date, type: Date

    field :motive_code, type: String

    common_indexes *%i[
      applicated_at
      effort_rate
      epci
      is_active
      max_date
      visit_conducted_date
      last_status_update_date
      lessor_event_informations
      sopra_code
      rank
      profile_score
      eligible_position
      applicated_position
      effort_rate
      als_id
      cotations
      requester_firstname
      requester_lastname
      requester_email
      requester_birth_date
      motive_code
      old_status
    ]

    index({housing_request_id: 1, status: 1}, {background: true})

    
    belongs_to :housing_offer, index: true, class_name: 'Hermes::HousingOffer'
    belongs_to :housing_request, index: true, class_name: 'Hermes::HousingRequest'
    
    validates_presence_of :housing_request_id, :housing_offer, unless: -> {disabled_or_deleted_status?}
    validates :status, inclusion: {in: Hermes::Statuses::HousingApplication.possible_status.values.flatten, allow_blank: true}
    validates :visit_status, inclusion: {in: Hermes::Statuses::HousingApplication.possible_status[:visit_status], allow_blank: true}
    validate :validate_request_supplement
    validate :can_create_application, on: :create
    validate :only_one_active_application_per_offer
    
    writer_auth_roles "collaborateur offre logement", 'superadministrateur'
    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    
    before_validation :before_validation_actions
    before_save :before_save_actions
    after_save :after_save_actions
    after_create :after_create_actions
    after_update :after_actions
    after_destroy :after_destroy_actions

    scope :selected_by_offer, ->(housing_offer_id) {
      where(housing_offer_id: housing_offer_id, status: "TRANSMIS BAILLEUR")
    }
    
    scope :actives, ->() {where(is_active: true)}
    
    HousingApplication.add_status(:single)
    HousingApplication.add_status(:single, "visit_status")
    
    ###############################
    #### Notification #############
    include ::Concerns::NotificationCenter::Notifiable

    def self.targets_selector(housing_app, opt = {})
      [housing_app.applicant].compact
    end
    
    def notification_mailer
      ::Hermes::HermesMailer
    end
    
    def notifiable_title(key, opt = {})
      I18n.t("#{key}.title")
    end
    
    def notifiable_body(key, opt = {})
      I18n.t("#{key}.body", number: housing_offer.reference, fullname: applicant.full_name)
    end
    
    def notifiable_path
      "https://als-hermes-int.be-idys.com/housing_applications/#{id}"
    end
    
    ###############################
    ###############################

    def validate_request_supplement
      hr = Hermes::HousingRequest.find(housing_request_id)
      return true unless self.new_record? && hr&.is_provider_imhoweb
      
      begin
        Timeout::timeout(10) do
          sleep(HERMES_CONFIG['sleeptimeout']) if HERMES_CONFIG['sleeptimeout'] # for testing
          hr.update_request_supplement(user, false, housing_offer.insee_code)
          housing_request.reload

          if hr.supplement_can_apply == false
            self.errors.add(:housing_request_id, "this housing_request is not allowed to apply")
            return false
          end
          
          return true
        end
      rescue Timeout::Error
        return true
      end
    end

    def can_create_application
      return if housing_request.can_still_apply

      errors.add :housing_request, 'Too many applications'
    end 

    def only_one_active_application_per_offer
      return unless (is_active_changed? && is_active) || housing_offer_id_changed? || housing_request_id_changed?

      if Hermes::HousingApplication.where(housing_offer_id: housing_offer_id, housing_request_id: housing_request_id, :id.ne => id, is_active: true).count > 0
        errors.add :housing_offer, 'Only one active application is allowed per offer'
      end
    end 

    def to_json(options = {})
      options[:except] ||= :sopra_code
      super
    end
    
    def as_json(options = {})
      options[:except] ||= :sopra_code
      super
    end
    
    def applicant
      housing_request&.main_applicant
    end
    
    def applicant_id
      applicant&.id
    end
    
    def update_offer_applicated_nb
      return unless housing_offer
      applicated_nb = Hermes::HousingApplication.where(
        is_active: true, 
        housing_offer_id: housing_offer_id,
      ).count
      housing_offer.set(updated_at: DateTime.current, applicated_nb: applicated_nb)
      housing_offer.handle_unpublished_status
      housing_offer.save if housing_offer.changed?
    end

    def update_request_attributes
      [
        :als_id,
        :requester_firstname,
        :requester_lastname,
        :requester_email,
        :requester_birth_date,
        :applicant_used_for_siret_firstname,
        :applicant_used_for_siret_lastname,
        :applicant_used_for_siret_email,
        :applicant_used_for_siret_birth_date,
      ].each do |att| 
        self[att] = housing_request && housing_request[att]
      end 
    end

    def als_id
      value = super
      return value if value.present?

      if housing_request
        value = housing_request.als_id
        set(als_id: value, updated_at: DateTime.current)
        self.als_id = value
        value
      end
    end

    def handle_status_changed
      applicant_cancel_application if disabled_or_deleted_status?

      if status&.upcase == 'RETENUE EN CAL'
        applicant.completed_requests_nb ? (applicant.completed_requests_nb += 1) : (applicant.completed_requests_nb = 1)
        applicant.save
      end
      create_public_share if status_changed? && status == 'TRANSMIS BAILLEUR'
    end 
      
    def send_updated_status_task
      return unless (code_to_send = calculate_code_to_send)

      send_update_status( 
        last_status_update_date: last_status_update_date, 
        date_event: date_event, 
        rank: rank, 
        code_to_send: code_to_send
      )
    end 

    def calculate_code_to_send(status_to_send = nil)
      code_to_send = nil

      if status_changed? || status_to_send
        code_to_send = Hermes::Statuses::HousingApplication.sopra_code(status_to_send || status)
      end

      if visit_status_changed? || (status_to_send && Hermes::Statuses::HousingApplication.possible_status[:visit_status].include?(status_to_send))
        case status_to_send || visit_status
        when 'VISITE PROPOSEE'
          code_to_send = "VISPRO"
        when 'VISITE EFFECTUEE'
          code_to_send = "VISEFF"
        end
      end

      code_to_send
    end 

    def send_update_status(status: nil, code_to_send: nil, last_status_update_date: nil , date_event: nil, rank: nil )
      code_to_send = calculate_code_to_send(status) unless code_to_send

      so = Hermes::BoClient::UpdateStatusTask.new({
        housing_application_id: id,
        housing_request_id: housing_request_id,
        housing_offer_id: housing_offer_id,
        code_evenement: code_to_send,
        date_evenement: last_status_update_date || Date.current,
        date_survenance_evenement: date_event,
        rang: rank 
      })
      so.execute
    end 
    
    def handle_visit_conducted_date
      if visit_conducted_date
        if visit_conducted_date.past?
          self.set(visit_status: 'VISITE EFFECTUEE', updated_at: DateTime.current) if visit_status.blank?
        else
          self.set(visit_status: 'VISITE PROPOSEE', updated_at: DateTime.current) if visit_status.blank?
        end
      end
    end

    def update_user_id_from_request
      self.user_id = housing_request.user_id
    end 
    
    def before_save_actions
      check_order_status if status_changed?
      update_right
      update_application_score
      update_cotations
      update_business_cotation
      update_user_id_from_request if housing_request_id_changed? && housing_request_id
      change_is_active if status_changed? || visit_status_changed?
      update_reader_writers_from_offer if housing_offer_id_changed?
      update_request_attributes if housing_request_id_changed?
      true
    end

    def check_order_status
      return if status_was.blank? || status.blank?

      if backward_status?(status, status_was)
        self.status = status_was
        self.date_event = date_event_was if date_event_was
        self.last_status_update_date = last_status_update_date_was if last_status_update_date_was
        self.rank = rank_was if rank_was
      end 
    end

    def has_a_indexed_status?
      !!Hermes::Statuses::HousingApplication.order_status[status] 
    end 

    def backward_status?(status, status_was)
      status_keys = Hermes::Statuses::HousingApplication.order_status.keys  
      index_status_w = status_keys.include?(status_was) && Hermes::Statuses::HousingApplication.order_status[status_was] 
      index_status = status_keys.include?(status) && Hermes::Statuses::HousingApplication.order_status[status]

      index_status_w && index_status && index_status <= index_status_w
    end 

    def update_application_score
      self.profile_score = housing_request.profile_score
    end 

    def update_business_cotation
      self.business_cotation = housing_request.business_cotation
    end 
    
    def update_reader_writers_from_offer
      old_offer = Hermes::HousingOffer.find(housing_offer_id_was) if housing_offer_id_was
      
      if housing_offer_id_was && old_offer
        self.readers = (self.readers || []) - (old_offer.readers || [])
        self.writers = (self.writers || []) - (old_offer.writers || [])
      end
      
      if housing_offer
        self.readers = (self.readers || []) + (housing_offer.readers || [])
        self.writers = (self.writers || []) + (housing_offer.writers || [])
      end
    end
    
    def change_is_active
      self.is_active = !inactive_status?
    end
    
    def before_validation_actions
      if status_changed?
        self.status_updated_at = DateTime.current
        self.sopra_code = Hermes::Statuses::HousingApplication.sopra_code(status)
        #  self.status = Hermes::Statuses::HousingApplication.correspondence(status || '')
      end
      
      if visit_status_changed?
        self.visit_status_updated_at = DateTime.current
        self.visit_status = I18n.transliterate(visit_status || '').upcase
      end
    end
    
    def reprise_update_right
      apps = Hermes::HousingApplication.where(is_active: true)
      apps.length
      apps.batch_size(5).each_with_index do |app, i|
        if app.housing_request
          app.update_right
          app.save
        end
      
      end
    end
    
    def update_right
      lessor_user_id = housing_offer&.lessor&.user_id&.to_s
      main_applicant = housing_request&.main_applicant
      reader_inst = [self, housing_request, main_applicant]
      if lessor_user_id && main_applicant
        if disabled_or_deleted_status?
          reader_inst.each do |inst|
            inst.pull(readers: lessor_user_id)
          end
          Hermes::Applicant.where(:id.in => housing_request.applicant_ids).pull(readers: lessor_user_id)
        else
          reader_inst.each do |inst|
            inst.add_to_set(readers: lessor_user_id)
          end
          Hermes::Applicant.where(:id.in => housing_request.applicant_ids).add_to_set(readers: lessor_user_id)
        end
        reader_inst.each do |inst|
          inst.add_to_set(readers: main_applicant.readers)
        end
      end
    end

    def update_request_application_info
      if is_active
        update_offer(:active_application_housing_offer_ids, :housing_offer_id, housing_offer_id.to_s)
        update_offer(:target_offer_reserving_businesses, :reserving_siren, housing_offer.reserving_business_name)
      else
        remove_offer(:active_application_housing_offer_ids)
        remove_offer(:target_offer_reserving_businesses) if status != 'LOGE'
      end
    end

    def update_offer(array_key, attribute_key, new_value)
      existing_entry = find_existing_entry(housing_request.send(array_key))
      if existing_entry
        existing_entry[attribute_key] = new_value
      else
        housing_request.send(array_key).push({ housing_application_id: id.to_s, attribute_key => new_value })
      end
    end
    
    def remove_offer(array_key)
      existing_entry = find_existing_entry(housing_request.send(array_key))
      housing_request.send(array_key).delete(existing_entry) if existing_entry
    end
    
    def find_existing_entry(entries)
      entries.to_a.find { |entry| entry[:housing_application_id].to_s == id.to_s }
    end
    
    def impact_offer_and_request_status
      if status == 'LOGE'
          housing_request.request_status = "SATISFAITE"
          housing_request.attributed_housing_district = housing_offer.district
          housing_request.attributed_housing_district_postal_code = housing_offer.postal_code || housing_offer.insee_code
      else
        housing_request.manage_entity_status
      end 
      
      case status
      when 'LOGE'
        housing_offer.update(
          offer_status: "LOGEMENT PLACE",
          lessor_status: "PLACE"
        )
      end
    end
    
    def after_save_actions
      check_status_changed if visit_status_changed? || status_changed?
      notify_applicant
      handle_status_changed if visit_status_changed? || status_changed? 
      send_updated_status_task if visit_status_changed? || status_changed?
      send_flux_50 if (status_changed? || visit_status_changed?) && disabled_or_deleted_status?
      update_request
      update_offer_applicated_nb if is_active_changed?
      cancel_other_applications if status_changed? && status == 'LOGE'
    end
    
    def after_actions
      handle_visit_conducted_date
    end

    def cancel_other_applications
      housing_request.dormant_application_ids.each do |app_id|
        app = Hermes::HousingApplication.find(app_id)
        next if app.has_a_indexed_status?

        app.update(is_active: false, status: 'ANNULEE')
      end
    end 

    def update_request
      update_request_application_info if is_active_changed?
      impact_offer_and_request_status if status_changed? || visit_status_changed?
      update_request_application_nb if status_changed?
      update_request_dormant_application_ids if rank_changed? || is_active_changed?

      housing_request.save if housing_request.changed?
    end 

    def update_request_dormant_application_ids
      is_active_application = housing_request.active_application_id == id.to_s
      is_dormant_application = housing_request.dormant_application_ids.include?(id.to_s)

      if [1, nil].include?(rank) && is_active 
        housing_request.dormant_application_ids.delete(id.to_s) if is_dormant_application
        housing_request.active_application_id = id.to_s
      elsif [2, 3].include?(rank) && is_active
        housing_request.dormant_application_ids.push(id.to_s) unless is_dormant_application
        housing_request.active_application_id = nil if is_active_application
      elsif !is_active
        housing_request.dormant_application_ids.delete(id.to_s) if is_dormant_application
        housing_request.active_application_id = nil if is_active_application
      end
    end 

    def update_request_application_nb
      statuses = [old_status.to_s, status.to_s]
      {
        'REFUSEE EN CAL': 'application_rejected_nb',
        'NON RETENUE': 'application_not_selected_nb',
        'REPRIS BAILLEUR': 'application_taken_back_nb',
      }.each do |k, v| 
        if statuses.include?(k.to_s)
          query = {housing_request: housing_request_id, status: k}
          query = query.merge(old_status: 'EN ATTENTE') if k.to_s == 'NON RETENUE'
          housing_request[v] = Hermes::HousingApplication.where(query).count(hint: "housing_request_id_1_status_1")
        end 
      end
    end 

    def update_cotations
      self.cotations = housing_request.cotations
    end 
    
    def after_create_actions
      housing_request.set(last_application_date: created_at, updated_at: DateTime.current)
      HandleRequestStateJob.set(wait: 1.days).perform_later(housing_request_id.to_s, 'request_application')
      
      if Rails.env == 'test'
        {exectuted: true, xml: to_xml}
      else
        Hermes::BoClient::ApplicationTask.new(
          {
            housing_application_id: id,
            housing_request_id: housing_request_id,
            housing_offer_id: housing_offer_id,
          }
        ).execute
      end
    end
    
    def after_destroy_actions
      # housing_request.with_lock(reload: false) do
        housing_request.save
      # end
      update_offer_applicated_nb
    end
    
    def disabled_or_deleted_status?
      Hermes::Statuses::HousingApplication.cancel_status.include?(self.status) || 
      Hermes::Statuses::HousingApplication.inactive_visit_status.include?(self.visit_status)
    end

    def inactive_status?
      Hermes::Statuses::HousingApplication.inactive_status.include?(self.status) || 
      Hermes::Statuses::HousingApplication.inactive_visit_status.include?(self.visit_status)
    end 

    def send_flux_50
      so = Hermes::BoClient::CancelStatusTask.new({
                                                    housing_application_id: id,
                                                    housing_offer_id: housing_offer_id,
                                                    housing_request_id: housing_request_id,
                                                    status_updated_at: status_updated_at || DateTime.now
                                                  })
      if Rails.env.test?
        #{exectuted: true, xml: so.to_xml}
        so.execute
        true
      else
        so.execute
      end
    end 
    
    def applicant_cancel_application
      housing_request.update_eligibility_query
      housing_request.save
    end
    
    def system_cancel_application!
      set(status: 'ANNULEE', is_active: false, updated_at: DateTime.current)
      housing_request.cancel_request
      housing_request.save
    end

    def create_public_share
      return nil unless housing_offer.lessor
      Hermes::HousingRequestShare.find_or_create_by(lessor_id: housing_offer.lessor.id, housing_request_id: housing_request.id)
    end
    
    def hr_offers_for_notification
      housing_request.eligibles_offers.where({
        reserved: false,
        offer_status:"LOGEMENT PUBLIE",
        lessor_status: "PUBLIE",
      }).order_by(date_publication_start: :desc).limit(6).pluck(:id).map(&:to_s)
    end

    def notify_applicant
      if status_changed?
        case status.upcase
        when 'RETENUE EN CAL'
          # notify lessor
          # notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_accepted])
        when 'LOGE'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_accepted], channels: ['email'])
        when 'TRANSMIS BAILLEUR'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_transferred], channels: ['email'])
        when 'NON RETENUE'
          housing_request.user.notify(
            key: NOTIFICATION_KEYS[:user][:application_not_selected],
            channels: ['email'],
            datas: {
              offer_ids: hr_offers_for_notification, 
              first_name: housing_request.requester_firstname,
              last_name: housing_request.requester_lastname , 
              external_ref: housing_offer.external_ref,
              district: housing_offer.district,
              applicated_at:(applicated_at || created_at)&.strftime("%d/%m/%Y")
            }
          )
        when 'REFUSEE EN CAL'
        housing_request.user.notify(
          key: NOTIFICATION_KEYS[:user][:application_refusal],
          channels: ['email'],
          datas: {
            offer_ids: hr_offers_for_notification, 
            first_name: housing_request.requester_firstname,
            last_name: housing_request.requester_lastname , 
            external_ref: housing_offer.external_ref,
            district: housing_offer.district,
            applicated_at:(applicated_at || created_at)&.strftime("%d/%m/%Y")
          }
          )
          # no_notification_motives = ['IV01', 'IV02', 'IV05']
          # unless no_notification_motives.include?(motive_code)
          # end
        when 'REPRIS BAILLEUR'
          housing_request.user.notify(
            key: NOTIFICATION_KEYS[:user][:application_offer_taken_back],
            channels: ['email'],
            datas: {
              offer_ids: hr_offers_for_notification, 
              first_name: housing_request.requester_firstname,
              last_name: housing_request.requester_lastname , 
              external_ref: housing_offer.external_ref,
              district: housing_offer.district,
              applicated_at:(applicated_at || created_at)&.strftime("%d/%m/%Y")
            }
          )
        when 'EN ATTENTE'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:pending], channels: [])
        when 'INSCRIPTION CAL'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:subscribed], channels: [])
        when 'DESISTEMENT SALARIE AVANT CAL', 'DESISTEMENT APRES CAL', 'DESISTEMENT'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:applicant_cancelled], channels: [])
        else
          nil
        end
      end
    end
  end
end
